@@ -1,9 +1,10 @@
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE TemplateHaskell #-}
 {- |
    Module      : Text.Pandoc.Readers.DocBook
-   Copyright   : Copyright (C) 2006-2022 John MacFarlane
+   Copyright   : Copyright (C) 2006-2023 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
@@ -13,7 +14,13 @@
 Conversion of DocBook XML to 'Pandoc' document.
 -}
 module Text.Pandoc.Readers.DocBook ( readDocBook ) where
+import Control.Monad (MonadPlus(mplus))
 import Control.Monad.State.Strict
+    ( MonadTrans(lift),
+      StateT(runStateT),
+      MonadState(get),
+      gets,
+      modify )
 import Data.ByteString (ByteString)
 import Data.FileEmbed
 import Data.Char (isSpace, isLetter, chr)
@@ -31,17 +38,18 @@ import Data.Text.Encoding (decodeUtf8)
 import qualified Data.Text as T
 import qualified Data.Text.Lazy as TL
 import Control.Monad.Except (throwError)
-import Text.HTML.TagSoup.Entity (lookupEntity)
+import Text.Pandoc.XML (lookupEntity)
 import Text.Pandoc.Error (PandocError(..))
 import Text.Pandoc.Builder
 import Text.Pandoc.Class.PandocMonad (PandocMonad, report)
 import Text.Pandoc.Options
 import Text.Pandoc.Logging (LogMessage(..))
-import Text.Pandoc.Shared (safeRead, extractSpaces)
+import Text.Pandoc.Shared (safeRead, extractSpaces, headerShift)
 import Text.Pandoc.Sources (ToSources(..), sourcesToText)
 import Text.TeXMath (readMathML, writeTeX)
 import qualified Data.Map as M
 import Text.Pandoc.XML.Light
+import Text.Pandoc.Walk (query)
 
 {-
 
@@ -313,7 +321,7 @@ List of all DocBook tags, with [x] indicating implemented,
 [x] para - A paragraph
 [ ] paramdef - Information about a function parameter in a programming language
 [x] parameter - A value or a symbolic reference to a value
-[ ] part - A division in a book
+[x] part - A division in a book
 [ ] partinfo - Meta-information for a Part
 [ ] partintro - An introduction to the contents of a part
 [ ] personblurb - A short description or note about a person
@@ -534,8 +542,6 @@ data DBState = DBState{ dbSectionLevel :: Int
                       , dbQuoteType    :: QuoteType
                       , dbMeta         :: Meta
                       , dbBook         :: Bool
-                      , dbFigureTitle  :: Inlines
-                      , dbFigureId     :: Text
                       , dbContent      :: [Content]
                       } deriving Show
 
@@ -544,8 +550,6 @@ instance Default DBState where
                , dbQuoteType = DoubleQuote
                , dbMeta = mempty
                , dbBook = False
-               , dbFigureTitle = mempty
-               , dbFigureId = mempty
                , dbContent = [] }
 
 
@@ -560,7 +564,14 @@ readDocBook _ inp = do
             docbookEntityMap
               (TL.fromStrict . handleInstructions . sourcesToText $ sources)
   (bs, st') <- flip runStateT (def{ dbContent = tree }) $ mapM parseBlock tree
-  return $ Pandoc (dbMeta st') (toList . mconcat $ bs)
+  let headerLevel (Header n _ _) = [n]
+      headerLevel _              = []
+  let bottomLevel = maybe 1 minimum $ nonEmpty $ query headerLevel bs
+  return $
+    -- handle the case where you have <part> or <chapter>
+    (if bottomLevel < 1
+        then headerShift (1 - bottomLevel)
+        else id) $ Pandoc (dbMeta st') $ toList $ mconcat bs
 
 -- We treat certain processing instructions by converting them to tags
 -- beginning "pi-".
@@ -582,11 +593,15 @@ getFigure e = do
   tit <- case filterChild (named "title") e of
               Just t  -> getInlines t
               Nothing -> return mempty
-  let ident = attrValue "id" e
-  modify $ \st -> st{ dbFigureTitle = tit, dbFigureId = ident }
-  res <- getBlocks e
-  modify $ \st -> st{ dbFigureTitle = mempty, dbFigureId = mempty }
-  return res
+  contents <- getBlocks e
+  let contents' =
+        case toList contents of
+          [Para [img@Image{}]] -> plain (fromList [img])
+          _ -> contents
+  return $ figureWith
+             (attrValue "id" e, [], [])
+             (simpleCaption $ plain tit)
+             contents'
 
 -- convenience function to get an attribute value, defaulting to ""
 attrValue :: Text -> Element -> Text
@@ -739,6 +754,8 @@ blockTags = Set.fromList $
   , "mediaobject"
   , "orderedlist"
   , "para"
+  , "part"
+  , "partinfo"
   , "preface"
   , "procedure"
   , "programlisting"
@@ -796,36 +813,35 @@ addToStart toadd bs =
 -- A DocBook mediaobject is a wrapper around a set of alternative presentations
 getMediaobject :: PandocMonad m => Element -> DB m Inlines
 getMediaobject e = do
-  figTitle <- gets dbFigureTitle
-  ident <- gets dbFigureId
-  (imageUrl, attr) <-
-    case filterElements (named "imageobject") e of
-      []  -> return (mempty, nullAttr)
-      (z:_) -> case filterChild (named "imagedata") z of
-                    Nothing -> return (mempty, nullAttr)
-                    Just i  -> let atVal a = attrValue a i
-                                   w = case atVal "width" of
-                                         "" -> []
-                                         d  -> [("width", d)]
-                                   h = case atVal "depth" of
-                                         "" -> []
-                                         d  -> [("height", d)]
-                                   id' = case atVal "id" of
-                                           x | T.null x  -> ident
-                                             | otherwise -> x
-                                   cs = T.words $ atVal "role"
-                                   atr = (id', cs, w ++ h)
-                               in  return (atVal "fileref", atr)
-  let getCaption el = case filterChild (\x -> named "caption" x
+  let (imageUrl, tit, attr) =
+        case filterElements (named "imageobject") e of
+          []  -> (mempty, mempty, nullAttr)
+          (z:_) ->
+            let tit' = maybe "" strContent $
+                         filterChild (named "objectinfo") z >>=
+                         filterChild (named "title")
+                (imageUrl', attr') =
+                  case filterChild (named "imagedata") z of
+                        Nothing -> (mempty, nullAttr)
+                        Just i  -> let atVal a = attrValue a i
+                                       w = case atVal "width" of
+                                             "" -> []
+                                             d  -> [("width", d)]
+                                       h = case atVal "depth" of
+                                             "" -> []
+                                             d  -> [("height", d)]
+                                       id' = atVal "id"
+                                       cs = T.words $ atVal "role"
+                                       atr = (id', cs, w ++ h)
+                                   in  (atVal "fileref", atr)
+            in  (imageUrl', tit', attr')
+  let capt = case filterChild (\x -> named "caption" x
                                             || named "textobject" x
-                                            || named "alt" x) el of
+                                            || named "alt" x) e of
                         Nothing -> return mempty
-                        Just z  -> mconcat <$>
+                        Just z  -> trimInlines . mconcat <$>
                                          mapM parseInline (elContent z)
-  let (capt, title) = if null figTitle
-                         then (getCaption e, "")
-                         else (return figTitle, "fig:")
-  fmap (imageWith attr imageUrl title) capt
+  fmap (imageWith attr imageUrl tit) capt
 
 getBlocks :: PandocMonad m => Element -> DB m Blocks
 getBlocks e =  mconcat <$>
@@ -876,6 +892,7 @@ parseBlock (Elem e) =
         "glosslist" -> definitionList <$>
                   mapM parseGlossEntry (filterChildren (named "glossentry") e)
         "chapter" -> modify (\st -> st{ dbBook = True}) >> sect 0
+        "part" -> modify (\st -> st{ dbBook = True}) >> sect (-1)
         "appendix" -> sect 0
         "preface" -> sect 0
         "bridgehead" -> para . strong <$> getInlines e
@@ -916,7 +933,8 @@ parseBlock (Elem e) =
           orderedListWith (start,listStyle,DefaultDelim) . handleCompact
             <$> listitems
         "variablelist" -> definitionList <$> deflistitems
-        "procedure" -> bulletList <$> steps
+        "procedure" -> orderedList <$> steps
+        "substeps" -> orderedList <$> steps
         "figure" -> getFigure e
         "informalfigure" -> getFigure e
         "mediaobject" -> para <$> getMediaobject e
@@ -934,6 +952,7 @@ parseBlock (Elem e) =
         "sect4info" -> skip  -- keywords & other metadata
         "sect5info" -> skip  -- keywords & other metadata
         "chapterinfo" -> skip -- keywords & other metadata
+        "partinfo" -> skip -- keywords & other metadata
         "glossaryinfo" -> skip  -- keywords & other metadata
         "appendixinfo" -> skip  -- keywords & other metadata
         "bookinfo" -> addMetadataFromElement e
@@ -1089,8 +1108,9 @@ parseBlock (Elem e) =
            -- the title: docbook references are ambiguous on that, and some implementations of admonitions
            -- (e.g. asciidoctor) just use an icon in all cases. To be conservative, we don't
            -- include the label and leave it to styling.
-           title <- case filterChild (named "title") e of
-                        Just t  -> divWith ("", ["title"], []) . plain <$> getInlines t
+           title <- divWith ("", ["title"], []) . plain <$>
+                    case filterChild (named "title") e of
+                        Just t  -> getInlines t
                         Nothing -> return mempty
            -- this will ignore the title element if it is present
            b <- getBlocks e
@@ -1166,7 +1186,7 @@ attrValueAsOptionalAttr n e = case attrValue n e of
 parseInline :: PandocMonad m => Content -> DB m Inlines
 parseInline (Text (CData _ s _)) = return $ text s
 parseInline (CRef ref) =
-  return $ text $ maybe (T.toUpper ref) T.pack $ lookupEntity (T.unpack ref)
+  return $ text $ fromMaybe (T.toUpper ref) $ lookupEntity ref
 parseInline (Elem e) =
   case qName (elName e) of
         "anchor" -> do
@@ -1265,7 +1285,10 @@ parseInline (Elem e) =
         "ulink" -> innerInlines (link (attrValue "url" e) "")
         "link" -> do
              ils <- innerInlines id
-             let href = case findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) e of
+             let href = case findAttrBy
+                               (\case
+                                 QName "href" _ _ -> True
+                                 _ -> False) e of
                                Just h -> h
                                _      -> "#" <> attrValue "linkend" e
              let ils' = if ils == mempty then str href else ils
@@ -1273,9 +1296,11 @@ parseInline (Elem e) =
              return $ linkWith attr href "" ils'
         "foreignphrase" -> innerInlines emph
         "emphasis" -> case attrValue "role" e of
+                             "bf"            -> innerInlines strong
                              "bold"          -> innerInlines strong
                              "strong"        -> innerInlines strong
                              "strikethrough" -> innerInlines strikeout
+                             "underline"     -> innerInlines underline
                              _               -> innerInlines emph
         "footnote" -> note . mconcat <$>
                          mapM parseBlock (elContent e)
@@ -1333,6 +1358,8 @@ parseInline (Elem e) =
          xrefTitleByElem el
              | not (T.null xrefLabel) = xrefLabel
              | otherwise              = case qName (elName el) of
+                  "book"         -> descendantContent "title" el
+                  "part"         -> descendantContent "title" el
                   "chapter"      -> descendantContent "title" el
                   "section"      -> descendantContent "title" el
                   "sect1"        -> descendantContent "title" el
@@ -1342,6 +1369,8 @@ parseInline (Elem e) =
                   "sect5"        -> descendantContent "title" el
                   "cmdsynopsis"  -> descendantContent "command" el
                   "funcsynopsis" -> descendantContent "function" el
+                  "figure"       -> descendantContent "title" el
+                  "table"        -> descendantContent "title" el
                   _              -> qName (elName el) <> "_title"
           where
             xrefLabel = attrValue "xreflabel" el
