@@ -105,8 +105,6 @@ inline'' = br
       <|> superscript
       <|> deleted
       <|> footnote
-      <|> inlineCode
-      <|> inlineFile
       <|> inlineRaw
       <|> math
       <|> autoLink
@@ -193,12 +191,6 @@ deleted = try $ B.strikeout <$> between (string "<del>") (try $ string "</del>")
 footnote :: PandocMonad m => DWParser m B.Inlines
 footnote = try $ B.note . B.para <$> between (string "((") (try $ string "))") nestedInlines
 
-inlineCode :: PandocMonad m => DWParser m B.Inlines
-inlineCode = codeTag B.codeWith "code"
-
-inlineFile :: PandocMonad m => DWParser m B.Inlines
-inlineFile = codeTag B.codeWith "file"
-
 inlineRaw :: PandocMonad m => DWParser m B.Inlines
 inlineRaw = try $ do
   char '<'
@@ -230,6 +222,7 @@ autoLink = try $ do
   state <- getState
   guard $ stateAllowLinks state
   (text, url) <- uri
+  guard $ not $ T.isInfixOf "%%//%%" text  -- see #9153
   guard $ checkLink (T.last url)
   return $ makeLink (text, url)
   where
@@ -247,7 +240,7 @@ str :: PandocMonad m => DWParser m B.Inlines
 str = B.str <$> (many1Char alphaNum <|> characterReference)
 
 symbol :: PandocMonad m => DWParser m B.Inlines
-symbol = B.str <$> countChar 1 nonspaceChar
+symbol = B.str <$> (notFollowedBy' blockCode *> countChar 1 nonspaceChar)
 
 link :: PandocMonad m => DWParser m B.Inlines
 link = try $ do
@@ -307,8 +300,11 @@ parseLink :: PandocMonad m
 parseLink f l r = f
   <$  textStr l
   <*> many1TillChar anyChar (lookAhead (void (char '|') <|> try (void $ textStr r)))
-  <*> optionMaybe (B.trimInlines . mconcat <$> (char '|' *> manyTill inline (try $ lookAhead $ textStr r)))
-  <*  textStr r
+  <*> ( (char '|' *> optionMaybe (B.trimInlines . B.text . T.pack <$>
+                       many1Till anyChar (lookAhead (try (textStr r)))))
+       <|> pure Nothing
+      )
+  <* textStr r
 
 -- | Split Interwiki link into left and right part
 -- | Return Nothing if it is not Interwiki link
@@ -414,7 +410,6 @@ blockElements = horizontalLine
             <|> indentedCode
             <|> quote
             <|> blockCode
-            <|> blockFile
             <|> blockRaw
             <|> table
 
@@ -448,8 +443,10 @@ parseList prefix marker =
   many1 ((<>) <$> item <*> fmap mconcat (many continuation))
   where
     continuation = try $ list ("  " <> prefix)
-    item = try $ textStr prefix *> char marker *> char ' ' *> itemContents
-    itemContents = B.plain . mconcat <$> many1Till inline' eol
+    item = try $ textStr prefix *> char marker *> char ' ' *>
+                   (mconcat <$> many1 itemContents <* eol)
+    itemContents = (B.plain . mconcat <$> many1 inline') <|>
+                   blockCode
 
 indentedCode :: PandocMonad m => DWParser m B.Blocks
 indentedCode = try $ B.codeBlock . T.unlines <$> many1 indentedLine
@@ -533,10 +530,8 @@ tableCell = try $ (second (B.plain . B.trimInlines . mconcat)) <$> cellContent
 
 
 blockCode :: PandocMonad m => DWParser m B.Blocks
-blockCode = codeTag B.codeBlockWith "code"
-
-blockFile :: PandocMonad m => DWParser m B.Blocks
-blockFile = codeTag B.codeBlockWith "file"
+blockCode = codeTag B.codeBlockWith "code" <|>
+            codeTag B.codeBlockWith "file"
 
 para :: PandocMonad m => DWParser m B.Blocks
 para = result . mconcat <$> many1Till inline endOfParaElement
@@ -544,7 +539,8 @@ para = result . mconcat <$> many1Till inline endOfParaElement
    endOfParaElement = lookAhead $ endOfInput <|> endOfPara <|> newBlockElement
    endOfInput       = try $ skipMany blankline >> skipSpaces >> eof
    endOfPara        = try $ blankline >> skipMany1 blankline
-   newBlockElement  = try $ blankline >> void blockElements
+   newBlockElement  = try (blankline >> void blockElements)
+                       <|> lookAhead (void blockCode)
    result content   = if F.all (==Space) content
                       then mempty
                       else B.para $ B.trimInlines content
