@@ -38,6 +38,7 @@ module Text.Pandoc.Shared (
                      -- * Date/time
                      normalizeDate,
                      -- * Pandoc block and inline list processing
+                     addPandocAttributes,
                      orderedListMarkers,
                      extractSpaces,
                      removeFormatting,
@@ -49,6 +50,7 @@ module Text.Pandoc.Shared (
                      linesToPara,
                      figureDiv,
                      makeSections,
+                     combineAttr,
                      uniqueIdent,
                      inlineListToIdentifier,
                      textToIdentifier,
@@ -94,8 +96,9 @@ import Data.Char (isAlpha, isLower, isSpace, isUpper, toLower, isAlphaNum,
 import Data.List (find, foldl', groupBy, intercalate, intersperse,
                   union, sortOn)
 import qualified Data.Map as M
-import Data.Maybe (mapMaybe, fromMaybe)
-import Data.Monoid (Any (..))
+import Data.Maybe (mapMaybe)
+import Data.Monoid (Any (..) )
+import Data.Semigroup (Min (..))
 import Data.Sequence (ViewL (..), ViewR (..), viewl, viewr)
 import qualified Data.Set as Set
 import qualified Data.Text as T
@@ -113,6 +116,9 @@ import Text.Pandoc.Extensions (Extensions, Extension(..), extensionEnabled)
 import Text.Pandoc.Generic (bottomUp)
 import Text.DocLayout (charWidth)
 import Text.Pandoc.Walk
+-- for addPandocAttributes:
+import Commonmark.Pandoc (Cm(..))
+import Commonmark (HasAttributes(..))
 
 --
 -- List processing
@@ -288,6 +294,16 @@ normalizeDate' s = fmap (formatTime defaultTimeLocale "%F")
 -- Pandoc block and inline list processing
 --
 
+-- | Add key-value attributes to a pandoc element. If the element
+-- does not have a slot for attributes, create an enclosing Span
+-- (for Inlines) or Div (for Blocks).  Note that both 'Cm () Inlines'
+-- and 'Cm () Blocks' are instances of 'HasAttributes'.
+addPandocAttributes
+  :: forall b . HasAttributes (Cm () b) => [(T.Text, T.Text)] -> b -> b
+addPandocAttributes [] bs = bs
+addPandocAttributes kvs bs =
+  unCm . addAttributes (("wrapper","1"):kvs) $ (Cm bs :: Cm () b)
+
 -- | Generate infinite lazy list of markers for an ordered list,
 -- depending on list attributes.
 orderedListMarkers :: (Int, ListNumberStyle, ListNumberDelim) -> [T.Text]
@@ -494,29 +510,28 @@ textToIdentifier exts =
 -- element a Header).  If the 'numbering' parameter is True, Header
 -- numbers are added via the number attribute on the header.
 -- If the baseLevel parameter is Just n, Header levels are
--- adjusted to be gapless starting at level n.
+-- adjusted so that the lowest header level is n.
+-- (There may still be gaps in header level if the author leaves them.)
 makeSections :: Bool -> Maybe Int -> [Block] -> [Block]
 makeSections numbering mbBaseLevel bs =
-  S.evalState (go bs) (mbBaseLevel, [])
+  S.evalState (go bs) []
  where
-  go :: [Block] -> S.State (Maybe Int, [Int]) [Block]
+  getLevel (Header level _ _) = Min level
+  getLevel _ = Min 99
+  minLevel = getMin $ query getLevel bs
+  go :: [Block] -> S.State [Int] [Block]
   go (Header level (ident,classes,kvs) title':xs) = do
-    (mbLevel, lastnum) <- S.get
-    let level' = fromMaybe level mbLevel
-    let lastnum' = take level' lastnum
-    let newnum =
-          if level' > 0
-             then case length lastnum' of
-                      x | "unnumbered" `elem` classes -> []
-                        | x >= level' -> init lastnum' ++ [last lastnum' + 1]
-                        | otherwise -> lastnum ++
-                             replicate (level' - length lastnum - 1) 0 ++ [1]
-             else []
-    unless (null newnum) $ S.modify $ \(mbl, _) -> (mbl, newnum)
+    lastnum <- S.get
+    let level' = maybe level (\n -> n + level - minLevel) mbBaseLevel
+    let adjustNum lev numComponent
+          | lev < level = numComponent
+          | lev == level = numComponent + 1
+          | otherwise = 0
+    let newnum = zipWith adjustNum [minLevel..level]
+                    (lastnum ++ repeat 0)
+    unless (null newnum || "unnumbered" `elem` classes) $ S.put newnum
     let (sectionContents, rest) = break (headerLtEq level) xs
-    S.modify $ \(_, ln) -> (fmap (+ 1) mbLevel, ln)
     sectionContents' <- go sectionContents
-    S.modify $ \(_, ln) -> (mbLevel, ln)
     rest' <- go rest
     let kvs' = -- don't touch number if already present
                case lookup "number" kvs of
@@ -550,13 +565,15 @@ makeSections numbering mbBaseLevel bs =
   go (x:xs) = (x :) <$> go xs
   go [] = return []
 
-  combineAttr :: Attr -> Attr -> Attr
-  combineAttr (id1, classes1, kvs1) (id2, classes2, kvs2) =
-    (if T.null id1 then id2 else id1,
-     nubOrd (classes1 ++ classes2),
-     foldr (\(k,v) kvs -> case lookup k kvs of
-                             Nothing -> (k,v):kvs
-                             Just _  -> kvs) mempty (kvs1 ++ kvs2))
+-- | Combine two 'Attr'. Classes are concatenated.  For the id and key-value
+-- attributes, the first one takes precedence in case of duplicates.
+combineAttr :: Attr -> Attr -> Attr
+combineAttr (id1, classes1, kvs1) (id2, classes2, kvs2) =
+  (if T.null id1 then id2 else id1,
+   nubOrd (classes1 ++ classes2),
+   foldr (\(k,v) kvs -> case lookup k kvs of
+                           Nothing -> (k,v):kvs
+                           Just _  -> kvs) kvs1 kvs2)
 
 headerLtEq :: Int -> Block -> Bool
 headerLtEq level (Header l _ _)  = l <= level
@@ -621,9 +638,13 @@ onlySimpleTableCells = all isSimpleCell . concat
 
 -- | Detect if a list is tight.
 isTightList :: [[Block]] -> Bool
-isTightList = all (\item -> firstIsPlain item || null item)
-  where firstIsPlain (Plain _ : _) = True
-        firstIsPlain _             = False
+isTightList = all isPlainItem
+  where
+    isPlainItem [] = True
+    isPlainItem (Plain _ : _) = True
+    isPlainItem [BulletList xs] = isTightList xs
+    isPlainItem [OrderedList _ xs] = isTightList xs
+    isPlainItem _ = False
 
 -- | Convert a list item containing tasklist syntax (e.g. @[x]@)
 -- to using @U+2610 BALLOT BOX@ or @U+2612 BALLOT BOX WITH X@.
@@ -758,7 +779,8 @@ formatCode attr = B.fromList . walk fmt . B.toList
 renderTags' :: [Tag T.Text] -> T.Text
 renderTags' = renderTagsOptions
                renderOptions{ optMinimize = matchTags ["hr", "br", "img",
-                                                       "meta", "link", "col", "use"]
+                                                       "meta", "link", "col",
+                                                       "use", "path", "rect"]
                             , optRawTag   = matchTags ["script", "style"] }
               where matchTags tags = flip elem tags . T.toLower
 
