@@ -87,6 +87,7 @@ import Data.String (fromString)
 data WriterState = WriterState
     { stNotes        :: [Html]  -- ^ List of notes
     , stEmittedNotes :: Int     -- ^ How many notes we've already pushed out to the HTML
+    , stEmittedNoteBlocks :: Int  -- ^ How many @\<div class=footnote>@ blocks we've already pushed out
     , stMath         :: Bool    -- ^ Math is used in document
     , stQuotes       :: Bool    -- ^ <q> tag is used
     , stHighlighting :: Bool    -- ^ Syntax highlighting is used
@@ -102,7 +103,11 @@ data WriterState = WriterState
     }
 
 defaultWriterState :: WriterState
-defaultWriterState = WriterState {stNotes= [], stEmittedNotes = 0, stMath = False, stQuotes = False,
+defaultWriterState = WriterState {stNotes= [],
+                                  stEmittedNotes = 0,
+                                  stEmittedNoteBlocks = 0,
+                                  stMath = False,
+                                  stQuotes = False,
                                   stHighlighting = False,
                                   stHtml5 = False,
                                   stEPUBVersion = Nothing,
@@ -530,6 +535,16 @@ footnoteSection opts refLocation startCounter notes = do
   let hrtag = if refLocation /= EndOfBlock
                  then (if html5 then H5.hr else H.hr) <> nl
                  else mempty
+  idName <- do
+    blockCount <- gets stEmittedNoteBlocks
+    modify $ \st -> st{ stEmittedNoteBlocks = blockCount + 1 }
+    return $
+      -- Keep the first note section's id undecorated to maintain a target for
+      -- old links which don't expect numbered sections, or for when the notes
+      -- are rendered all together at the end of the document.
+      if blockCount <= 0
+        then "footnotes"
+        else "footnotes-" <> show (blockCount + 1)
   let additionalClassName = case refLocation of
         EndOfBlock -> "footnotes-end-of-block"
         EndOfDocument -> "footnotes-end-of-document"
@@ -539,19 +554,19 @@ footnoteSection opts refLocation startCounter notes = do
   let container x
         | html5
         , epubVersion == Just EPUB3
-                = H5.section ! A.id "footnotes"
+                = H5.section ! A.id (fromString idName)
                              ! A.class_ className
                              ! customAttribute "epub:type" "footnotes" $ x
         | html5
         , refLocation == EndOfDocument
-        , slideVariant == RevealJsSlides -- need a section for a new slide:
-                = H5.section ! A5.id "footnotes"
+        -- Note: we need a section for a new slide in slide formats.
+                = H5.section ! A5.id (fromString idName)
                              ! A5.class_ className
                              ! A5.role "doc-endnotes"
                              $ x
-        | html5 = H5.aside   ! prefixedId opts "footnotes"
+        | html5 = H5.aside   ! prefixedId opts (fromString idName)
                              ! A5.class_ className
-                             ! A5.role "doc-endnotes"
+                             ! A5.role "doc-footnote"
                              $ x
         | slideVariant /= NoSlides = H.div ! A.class_ "footnotes slide" $ x
         | otherwise = H.div ! A.class_ className $ x
@@ -721,17 +736,22 @@ adjustNumbers opts doc =
      then doc
      else walk go doc
   where
-   go (Div (ident,"section":classes,kvs) lst) =
-     Div (ident,"section":classes,map fixnum kvs) lst
+   go (Div (ident,"section":classes,kvs) lst@(Header level _ _ : _)) =
+     Div (ident,"section":classes,map (fixnum level) kvs) lst
    go (Header level (ident,classes,kvs) lst) =
-     Header level (ident,classes,map fixnum kvs) lst
+     Header level (ident,classes,map (fixnum level) kvs) lst
    go x = x
-   fixnum ("number",num) = ("number",
+   fixnum level ("number",num) = ("number",
                                showSecNum $ zipWith (+)
                                (writerNumberOffset opts ++ repeat 0)
-                               (map (fromMaybe 0 . safeRead) $
+                               (padTo level $
+                                map (fromMaybe 0 . safeRead) $
                                 T.split (=='.') num))
-   fixnum x = x
+   fixnum _ x = x
+   padTo n xs =
+     case n - length xs of
+       x | x > 0 -> replicate x 0 ++ xs
+         | otherwise -> xs
    showSecNum = T.intercalate "." . map tshow
 
 blockToHtmlInner :: PandocMonad m => WriterOptions -> Block -> StateT WriterState m Html
@@ -1552,6 +1572,7 @@ inlineToHtml opts inline = do
                                     then link'
                                     else link' ! A.title (toValue tit)
     (Image attr@(_, _, attrList) txt (s, tit)) -> do
+                        epubVersion <- gets stEPUBVersion
                         let alternate = stringify txt
                         slideVariant <- gets stSlideVariant
                         let isReveal = slideVariant == RevealJsSlides
@@ -1564,8 +1585,9 @@ inlineToHtml opts inline = do
                               [A.title $ toValue tit | not (T.null tit)] ++
                               attrs
                             imageTag = (if html5 then H5.img else H.img
-                              , [A.alt $ toValue alternate | not (null txt) &&
-                                  isNothing (lookup "alt" attrList)] )
+                              , [A.alt $ toValue alternate |
+                                  isNothing (lookup "alt" attrList) &&
+                                  (isJust epubVersion || not (null txt))] )
                             mediaTag tg fallbackTxt =
                               let linkTxt = if null txt
                                             then fallbackTxt
@@ -1603,7 +1625,7 @@ inlineToHtml opts inline = do
                                              else H.sup)
                                        $ toHtml ref
                         return $ case epubVersion of
-                                      Just EPUB3 -> link ! customAttribute "epub:type" "noteref"
+                                      Just EPUB3 -> link ! customAttribute "epub:type" "noteref" ! customAttribute "role" "doc-noteref"
                                       _ | html5  -> link ! A5.role "doc-noteref"
                                       _          -> link
     (Cite cits il)-> do contents <- inlineListToHtml opts
@@ -1666,7 +1688,7 @@ blockListToNote opts ref blocks = do
              _ -> Para [backlink , Str "."] : blocks
       contents <- blockListToHtml opts blocks'
       let noteItem = (if epubv == EPUB3
-                         then H5.aside ! customAttribute "epub:type" "footnote"
+                         then H5.aside ! customAttribute "epub:type" "footnote" ! customAttribute "role" "doc-footnote"
                          else H.div) ! prefixedId opts ("fn" <> ref)
                       $ nl >> contents >> nl
       return $ noteItem >> nl
@@ -1731,11 +1753,20 @@ intrinsicEventsHTML4 =
   [ "onclick", "ondblclick", "onmousedown", "onmouseup", "onmouseover"
   , "onmouseout", "onmouseout", "onkeypress", "onkeydown", "onkeyup"]
 
+
+-- | Check to see if Format is valid HTML
 isRawHtml :: PandocMonad m => Format -> StateT WriterState m Bool
 isRawHtml f = do
   html5 <- gets stHtml5
   return $ f == Format "html" ||
-           ((html5 && f == Format "html5") || f == Format "html4")
+           ((html5 && f == Format "html5") || f == Format "html4") ||
+           isSlideVariant f
+
+-- | Check to see if Format matches with an HTML slide variant
+isSlideVariant :: Format -> Bool
+isSlideVariant f = f `elem` [Format "s5", Format "slidy", Format "slideous",
+                             Format "dzslides", Format "revealjs"]
+
 
 -- We need to remove links from link text, because an <a> element is
 -- not allowed inside another <a> element.
