@@ -65,11 +65,43 @@ import Text.Pandoc.Writers.LaTeX.Util (stringToLaTeX, StringContext(..),
 import Text.Pandoc.Writers.Shared
 import qualified Text.Pandoc.Writers.AnnotatedTable as Ann
 
+-- Work around problems with notes inside emphasis (see #8982)
+isolateBigNotes :: ([Inline] -> Inline) -> [Inline] -> [Inline]
+isolateBigNotes constructor xs =
+  let (before, after) = break isBigNote xs
+  in case after of
+       (noteInline:rest) -> constructor before :
+                            noteInline :
+                            isolateBigNotes constructor rest
+       [] -> [constructor xs]
+
+isBigNote :: Inline -> Bool
+isBigNote (Note [Plain _]) = False  -- A small note
+isBigNote (Note [Para _]) =  False  -- A small note
+isBigNote (Note _) = True  -- A big note
+isBigNote _ = False  -- Not a note
+
+raiseBigNotes :: [Inline] -> [Inline]
+raiseBigNotes (Emph inner : xs)
+  = isolateBigNotes Emph (raiseBigNotes inner) ++ raiseBigNotes xs
+raiseBigNotes (Strong inner : xs)
+  = isolateBigNotes Strong (raiseBigNotes inner) ++ raiseBigNotes xs
+raiseBigNotes (Underline inner : xs)
+  = isolateBigNotes Underline (raiseBigNotes inner) ++ raiseBigNotes xs
+raiseBigNotes (Strikeout inner : xs)
+  = isolateBigNotes Strikeout (raiseBigNotes inner) ++ raiseBigNotes xs
+raiseBigNotes (x : xs)          = x : raiseBigNotes xs
+raiseBigNotes [] = []
+
 -- | Convert Pandoc to LaTeX.
 writeLaTeX :: PandocMonad m => WriterOptions -> Pandoc -> m Text
-writeLaTeX options document =
-  evalStateT (pandocToLaTeX options document) $
-    startingState options
+writeLaTeX options document = do
+  let Any hasBigNotes =
+       query (\il -> if isBigNote il then Any True else Any False) document
+  let document' = if hasBigNotes
+                     then walk raiseBigNotes document
+                     else document
+  evalStateT (pandocToLaTeX options document') $ startingState options
 
 -- | Convert Pandoc to LaTeX Beamer.
 writeBeamer :: PandocMonad m => WriterOptions -> Pandoc -> m Text
@@ -95,6 +127,13 @@ pandocToLaTeX options (Pandoc meta blocks) = do
   let colwidth = if writerWrapText options == WrapAuto
                     then Just $ writerColumns options
                     else Nothing
+  docLangs <- catMaybes <$>
+      mapM (toLang . Just) (nubOrd (query (extract "lang") blocks))
+  mblang <- toLang $ case getLang options meta of
+                          Just l -> Just l
+                          Nothing | null docLangs -> Nothing
+                                  | otherwise     -> Just "en"
+  modify $ \s -> s{ stLang = mblang }
   metadata <- metaToContext options
               blockListToLaTeX
               (fmap chomp . inlineListToLaTeX)
@@ -135,8 +174,7 @@ pandocToLaTeX options (Pandoc meta blocks) = do
   st <- get
   titleMeta <- stringToLaTeX TextString $ stringify $ docTitle meta
   authorsMeta <- mapM (stringToLaTeX TextString . stringify) $ docAuthors meta
-  docLangs <- catMaybes <$>
-      mapM (toLang . Just) (nubOrd (query (extract "lang") blocks))
+  -- we need a default here since lang is used in template conditionals
   let hasStringValue x = isJust (getField x metadata :: Maybe (Doc Text))
   let geometryFromMargins = mconcat $ intersperse ("," :: Doc Text) $
                             mapMaybe (\(x,y) ->
@@ -146,12 +184,6 @@ pandocToLaTeX options (Pandoc meta blocks) = do
                               ,("tmargin","margin-top")
                               ,("bmargin","margin-bottom")
                               ]
-  mblang <- toLang $ case getLang options meta of
-                          Just l -> Just l
-                          Nothing | null docLangs -> Nothing
-                                  | otherwise     -> Just "en"
-  -- we need a default here since lang is used in template conditionals
-  let otherLangs = [l | l <- docLangs, mblang /= Just l]
 
   let dirs = query (extract "dir") blocks
 
@@ -218,15 +250,18 @@ pandocToLaTeX options (Pandoc meta blocks) = do
                           -> resetField "papersize" ("a" <> ds)
                       _   -> id)
                   metadata
+  let babelLang = mblang >>= toBabel
   let context' =
           -- note: lang is used in some conditionals in the template,
           -- so we need to set it if we have any babel/polyglossia:
           maybe id (\l -> defField "lang"
                       (literal $ renderLang l)) mblang
         $ maybe id (\l -> defField "babel-lang"
-                      (literal l)) (mblang >>= toBabel)
+                      (literal l)) babelLang
         $ defField "babel-otherlangs"
-             (map literal $ mapMaybe toBabel otherLangs)
+             (map literal
+               (nubOrd . catMaybes . filter (/= babelLang)
+                $ map toBabel docLangs))
         $ defField "latex-dir-rtl"
            ((render Nothing <$> getField "dir" context) ==
                Just ("rtl" :: Text)) context
@@ -371,7 +406,13 @@ blockToLaTeX (Para [Str ".",Space,Str ".",Space,Str "."]) = do
      then blockToLaTeX (RawBlock "latex" "\\pause")
      else inlineListToLaTeX [Str ".",Space,Str ".",Space,Str "."]
 blockToLaTeX (Para lst) =
-  inlineListToLaTeX lst
+  if null lst
+     then do
+       opts <- gets stOptions
+       if isEnabled Ext_empty_paragraphs opts
+          then pure "\\hfill\\par"
+          else pure mempty
+     else inlineListToLaTeX lst
 blockToLaTeX (LineBlock lns) =
   blockToLaTeX $ linesToPara lns
 blockToLaTeX (BlockQuote lst) = do
@@ -742,24 +783,37 @@ inlineListToLaTeX :: PandocMonad m
                   => [Inline]  -- ^ Inlines to convert
                   -> LW m (Doc Text)
 inlineListToLaTeX lst = hcat <$>
-  mapM inlineToLaTeX (fixLineInitialSpaces . fixInitialLineBreaks $ lst)
+  mapM inlineToLaTeX
+    (addKerns . fixLineInitialSpaces . fixInitialLineBreaks $ lst)
     -- nonbreaking spaces (~) in LaTeX don't work after line breaks,
-    -- so we turn nbsps after hard breaks to \hspace commands.
-    -- this is mostly used in verse.
+    -- so we insert a strut: this is mostly used in verse.
  where fixLineInitialSpaces [] = []
        fixLineInitialSpaces (LineBreak : Str s : xs)
          | Just ('\160', _) <- T.uncons s
-         = LineBreak : fixNbsps s <> fixLineInitialSpaces xs
+         = LineBreak : RawInline "latex" "\\strut " : Str s
+            : fixLineInitialSpaces xs
        fixLineInitialSpaces (x:xs) = x : fixLineInitialSpaces xs
-       fixNbsps s = let (ys,zs) = T.span (=='\160') s
-                    in  replicate (T.length ys) hspace <> [Str zs]
-       hspace = RawInline "latex" "\\hspace*{0.333em}"
        -- We need \hfill\break for a line break at the start
        -- of a paragraph. See #5591.
        fixInitialLineBreaks (LineBreak:xs) =
          RawInline (Format "latex") "\\hfill\\break\n" :
            fixInitialLineBreaks xs
        fixInitialLineBreaks xs = xs
+       addKerns [] = []
+       addKerns (Str s : q@Quoted{} : rest)
+         | isQuote (T.takeEnd 1 s) =
+           Str s : RawInline (Format "latex") "\\," : addKerns (q:rest)
+       addKerns (q@Quoted{} : Str s : rest)
+         | isQuote (T.take 1 s) =
+           q : RawInline (Format "latex") "\\," : addKerns (Str s : rest)
+       addKerns (x:xs) = x : addKerns xs
+       isQuote "\"" = True
+       isQuote "'" = True
+       isQuote "\x2018" = True
+       isQuote "\x2019" = True
+       isQuote "\x201C" = True
+       isQuote "\x201D" = True
+       isQuote _ = False
 
 -- | Convert inline element to LaTeX
 inlineToLaTeX :: PandocMonad m
@@ -814,7 +868,10 @@ inlineToLaTeX (Strikeout lst) = do
   -- incorrect results if there is a space, see #5529
   contents <- inlineListToLaTeX $ walk (concatMap protectCode) lst
   modify $ \s -> s{ stStrikeout = True }
-  return $ inCmd "st" contents
+  -- soul doesn't like \(..\) delimiters, so we change these to $ (#9597):
+  let fixMath = T.replace "\\(" "$" . T.replace "\\)" "$" .
+                T.replace "\\[" "$$" . T.replace "\\]" "$$"
+  return $ inCmd "st" $ fixMath <$> contents
 inlineToLaTeX (Superscript lst) =
   inCmd "textsuperscript" <$> inlineListToLaTeX lst
 inlineToLaTeX (Subscript lst) =
@@ -898,13 +955,21 @@ inlineToLaTeX (Quoted qt lst) = do
                DoubleQuote -> "\\enquote" <> braces contents
                SingleQuote -> "\\enquote*" <> braces contents
      else do
-       let s1 = if not (null lst) && isQuoted (head lst)
-                   then "\\,"
-                   else empty
-       let s2 = if not (null lst) && isQuoted (last lst)
-                   then "\\,"
-                   else empty
-       let inner = s1 <> contents <> s2
+       let endsWithQuote xs =
+             case reverse xs of
+                   Quoted{}:_ -> True
+                   Span _ ys : _ -> endsWithQuote ys
+                   Str s:_ -> T.takeEnd 1 s == "'"
+                   _ -> False
+       let beginsWithQuote xs =
+             case xs of
+                   Quoted{}:_ -> True
+                   Span _ ys : _ -> beginsWithQuote ys
+                   Str s:_ -> T.take 1 s == "`"
+                   _ -> False
+       let inner = (if beginsWithQuote lst then "\\," else mempty)
+                    <> contents
+                    <> (if endsWithQuote lst then "\\," else mempty)
        return $ case qt of
                 DoubleQuote ->
                    if isEnabled Ext_smart opts
@@ -914,10 +979,6 @@ inlineToLaTeX (Quoted qt lst) = do
                    if isEnabled Ext_smart opts
                       then char '`' <> inner <> char '\''
                       else char '\x2018' <> inner <> char '\x2019'
-    where
-      isQuoted (Span _ (x:_)) = isQuoted x
-      isQuoted (Quoted _ _)   = True
-      isQuoted _              = False
 inlineToLaTeX (Str str) = do
   setEmptyLine False
   liftM literal $ stringToLaTeX TextString str
