@@ -368,39 +368,58 @@ blockToMarkdown' :: PandocMonad m
                  => WriterOptions -- ^ Options
                  -> Block         -- ^ Block element
                  -> MD m (Doc Text)
-blockToMarkdown' opts (Div attrs ils) = do
-  contents <- blockListToMarkdown opts ils
-  variant <- asks envVariant
-  return $
-     case () of
-         _ | variant == Markua ->
-                   case () of
-                        () | "blurb" `elem` classes' -> prefixed "B> " contents <> blankline
-                           | "aside" `elem` classes' -> prefixed "A> " contents <> blankline
-                           -- necessary to enable option to create a bibliography
-                           | (take 3 (T.unpack id')) == "ref" -> contents <> blankline
-                           | otherwise -> contents <> blankline
-           | isEnabled Ext_fenced_divs opts &&
-             attrs /= nullAttr ->
-                let attrsToMd = if variant == Commonmark
-                                then attrsToMarkdown opts
-                                else classOrAttrsToMarkdown opts
-                in nowrap (literal ":::" <+> attrsToMd attrs) $$
-                   chomp contents $$
-                   literal ":::" <> blankline
-           | isEnabled Ext_native_divs opts ||
-             (isEnabled Ext_raw_html opts &&
-              (variant == Commonmark ||
-               isEnabled Ext_markdown_in_html_blocks opts)) ->
-                tagWithAttrs "div" attrs <> blankline <>
-                contents <> blankline <> "</div>" <> blankline
-           | isEnabled Ext_raw_html opts &&
-             isEnabled Ext_markdown_attribute opts ->
-                tagWithAttrs "div" attrs' <> blankline <>
-                contents <> blankline <> "</div>" <> blankline
-           | otherwise -> contents <> blankline
-       where (id',classes',kvs') = attrs
-             attrs' = (id',classes',("markdown","1"):kvs')
+blockToMarkdown' opts (Div attrs@(_,classes,_) bs)
+  | isEnabled Ext_alerts opts
+  , (cls:_) <- classes
+  , cls `elem` ["note", "tip", "warning", "caution", "important"]
+  , (Div ("", ["title"], []) _ : Para ils : bs') <- bs
+   = blockToMarkdown' opts $ BlockQuote $
+       (Para (RawInline (Format "markdown") (case cls of
+         "note" -> "[!NOTE]\n"
+         "tip" -> "[!TIP]\n"
+         "warning" -> "[!WARNING]\n"
+         "caution" -> "[!CAUTION]\n"
+         "important" -> "[!IMPORTANT]\n"
+         _ -> "[!NOTE]\n") : ils)) : bs'
+  | otherwise = do
+    contents <- blockListToMarkdown opts bs
+    variant <- asks envVariant
+    return $
+       case () of
+           _ | variant == Markua ->
+                 case () of
+                      () | "blurb" `elem` classes'
+                           -> prefixed "B> " contents <> blankline
+                         | "aside" `elem` classes'
+                           -> prefixed "A> " contents <> blankline
+                         -- necessary to enable option to create a bibliography
+                         | (take 3 (T.unpack id')) == "ref"
+                           -> contents <> blankline
+                         | otherwise -> contents <> blankline
+             | isEnabled Ext_fenced_divs opts &&
+               attrs /= nullAttr ->
+                  let attrsToMd = if variant == Commonmark
+                                  then attrsToMarkdown opts
+                                  else classOrAttrsToMarkdown opts
+                      divNesting = computeDivNestingLevel bs
+                      numcolons = 3 + divNesting
+                      colons = literal $ T.replicate numcolons ":"
+                  in nowrap (colons <+> attrsToMd attrs) $$
+                     chomp contents $$
+                     colons <> blankline
+             | isEnabled Ext_native_divs opts ||
+               (isEnabled Ext_raw_html opts &&
+                (variant == Commonmark ||
+                 isEnabled Ext_markdown_in_html_blocks opts)) ->
+                  tagWithAttrs "div" attrs <> blankline <>
+                  contents <> blankline <> "</div>" <> blankline
+             | isEnabled Ext_raw_html opts &&
+               isEnabled Ext_markdown_attribute opts ->
+                  tagWithAttrs "div" attrs' <> blankline <>
+                  contents <> blankline <> "</div>" <> blankline
+             | otherwise -> contents <> blankline
+         where (id',classes',kvs') = attrs
+               attrs' = (id',classes',("markdown","1"):kvs')
 blockToMarkdown' opts (Plain inlines) = do
   -- escape if para starts with ordered list marker
   variant <- asks envVariant
@@ -574,16 +593,19 @@ blockToMarkdown' opts (BlockQuote blocks) = do
         | otherwise            = "> "
   contents <- blockListToMarkdown opts blocks
   return $ prefixed leader contents <> blankline
-blockToMarkdown' opts t@(Table _ blkCapt specs thead tbody tfoot) = do
+blockToMarkdown' opts t@(Table (ident,_,_) blkCapt specs thead tbody tfoot) = do
   let (caption, aligns, widths, headers, rows) = toLegacyTable blkCapt specs thead tbody tfoot
   let numcols = maximum (length aligns :| length widths :
                            map length (headers:rows))
   caption' <- inlineListToMarkdown opts caption
-  let caption''
+  let caption'' = if T.null ident
+                     then caption'
+                     else caption' <+> attrsToMarkdown opts (ident,[],[])
+  let caption'''
         | null caption = blankline
         | isEnabled Ext_table_captions opts
-        = blankline $$ (": " <> caption') $$ blankline
-        | otherwise = blankline $$ caption' $$ blankline
+        = blankline $$ (": " <> caption'') $$ blankline
+        | otherwise = blankline $$ caption'' $$ blankline
   let hasSimpleCells = onlySimpleTableCells $ headers : rows
   let isSimple = hasSimpleCells && all (==0) widths
   let isPlainBlock (Plain _) = True
@@ -636,7 +658,7 @@ blockToMarkdown' opts t@(Table _ blkCapt specs thead tbody tfoot) = do
                    literal . removeBlankLinesInHTML <$>
                    writeHtml5String opts{ writerTemplate = Nothing } (Pandoc nullMeta [t])
             | otherwise -> return (id, literal "[TABLE]")
-  return $ nst (tbl $$ caption'') $$ blankline
+  return $ nst (tbl $$ caption''') $$ blankline
 blockToMarkdown' opts (BulletList items) = do
   contents <- inList $ mapM (bulletListItemToMarkdown opts) items
   return $ (if isTightList items then vcat else vsep)
@@ -876,3 +898,9 @@ removeBlankLinesInHTML = T.pack . go False . T.unpack
         go !afternewline (!c:cs)
           | isSpace c = c : go afternewline cs
           | otherwise = c : go False cs
+
+computeDivNestingLevel :: [Block] -> Int
+computeDivNestingLevel = foldr go 0
+ where
+   go (Div _ bls') n = max (n + 1) (foldr go (n + 1) bls')
+   go _ n = n
